==> fosslight_dependency-3.14.3.tar.gz: fosslight_dependency-3.14.3/src/fosslight_dependency/package_manager/Npm.py <==
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 LG Electronics Inc.
# SPDX-License-Identifier: Apache-2.0

import os
import logging
import subprocess
import json
import shutil
import re
import fosslight_util.constant as constant
import fosslight_dependency.constant as const
from fosslight_dependency._package_manager import PackageManager, get_url_to_purl

logger = logging.getLogger(constant.LOGGER_NAME)
node_modules = 'node_modules'


class Npm(PackageManager):
    package_manager_name = const.NPM

    dn_url = 'https://www.npmjs.com/package/'
    input_file_name = 'tmp_npm_license_output.json'
    tmp_custom_json = 'custom.json'
    flag_tmp_node_modules = False

    def __init__(self, input_dir, output_dir):
        super().__init__(self.package_manager_name, self.dn_url, input_dir, output_dir)

    def __del__(self):
        if os.path.isfile(self.input_file_name):
            os.remove(self.input_file_name)
        if self.flag_tmp_node_modules:
            shutil.rmtree(node_modules, ignore_errors=True)
        if os.path.exists(self.tmp_custom_json):
            os.remove(self.tmp_custom_json)

    def run_plugin(self):
        ret = self.start_license_checker()
        return ret

    def start_license_checker(self):
        ret = True
        license_checker_cmd = f'license-checker --production --json --out {self.input_file_name}'
        custom_path_option = ' --customPath '
        npm_install_cmd = 'npm install --production'

        if os.path.isdir(node_modules) != 1:
            logger.info(f"node_modules directory is not existed. So it executes '{npm_install_cmd}'.")
            self.flag_tmp_node_modules = True
            cmd_ret = subprocess.call(npm_install_cmd, shell=True)
            if cmd_ret != 0:
                logger.error(f"{npm_install_cmd} returns an error")
                return False

        # customized json file for obtaining specific items with license-checker
        self.make_custom_json(self.tmp_custom_json)

        cmd = license_checker_cmd + custom_path_option + self.tmp_custom_json
        cmd_ret = subprocess.call(cmd, shell=True)
        if cmd_ret != 0:
            logger.error(f"It returns the error: {cmd}")
            logger.error("Please check if the license-checker is installed.(sudo npm install -g license-checker)")
            ret = False
        else:
            self.append_input_package_list_file(self.input_file_name)
        if os.path.exists(self.tmp_custom_json):
            os.remove(self.tmp_custom_json)

        return ret

    def make_custom_json(self, tmp_custom_json):
        with open(tmp_custom_json, 'w', encoding='utf8') as custom:
            custom.write(
                "{\n\t\"name\": \"\",\n\t\"version\": \"\",\n\t\"licenses\": \"\",\n\t\"repository\": \
                \"\",\n\t\"url\": \"\",\n\t\"copyright\": \"\",\n\t\"licenseText\": \"\"\n}\n".encode().decode("utf-8"))

    def parse_rel_dependencies(self, rel_name, rel_ver, rel_dependencies):
        _dependencies = 'dependencies'
        _version = 'version'
        _peer = 'peerMissing'

        for rel_dep_name in rel_dependencies.keys():
            # Optional, non-installed dependencies are listed as empty objects
            if rel_dependencies[rel_dep_name] == {}:
                continue
            if _peer in rel_dependencies[rel_dep_name]:
                if rel_dependencies[rel_dep_name][_peer]:
                    continue
            if f'{rel_name}({rel_ver})' not in self.relation_tree:
                self.relation_tree[f'{rel_name}({rel_ver})'] = []
            elif f'{rel_dep_name}({rel_dependencies[rel_dep_name][_version]})' in self.relation_tree[f'{rel_name}({rel_ver})']:
                continue
            self.relation_tree[f'{rel_name}({rel_ver})'].append(f'{rel_dep_name}({rel_dependencies[rel_dep_name][_version]})')
            if _dependencies in rel_dependencies[rel_dep_name]:
                self.parse_rel_dependencies(rel_dep_name, rel_dependencies[rel_dep_name][_version],
                                            rel_dependencies[rel_dep_name][_dependencies])

    def parse_transitive_relationship(self):
        _dependencies = 'dependencies'
        _version = 'version'
        _name = 'name'
        ret = True
        err_msg = ''

        cmd = 'npm ls -a --omit=dev --json -s'
        result = subprocess.run(cmd, shell=True, capture_output=True, text=True, encoding='utf-8')
        rel_tree = result.stdout
        if rel_tree is None:
            logger.error(f"It returns the error: {cmd}")
            logger.error(f"No output for {cmd}")
            ret = False
        if ret:
            if result.returncode == 1:
                logger.warning(f"'{cmd}' returns error code: {result.stderr}")

            try:
                rel_json = json.loads(rel_tree)
                if len(rel_json) < 1:
                    ret = False
                else:
                    self.package_name = f'{rel_json[_name]}({rel_json[_version]})'
                    if _dependencies in rel_json:
                        self.parse_rel_dependencies(rel_json[_name], rel_json[_version], rel_json[_dependencies])
            except Exception as e:
                ret = False
                err_msg = e
        return ret, err_msg

    def parse_direct_dependencies(self):
        if not self.direct_dep:
            return
        try:
            if os.path.isfile(const.SUPPORT_PACKAE.get(self.package_manager_name)):
                ret, err_msg = self.parse_transitive_relationship()
                if not ret:
                    self.direct_dep = False
                    logger.warning(f'It cannot print direct/transitive dependency: {err_msg}')
            else:
                logger.info('Direct/transitive support is not possible because the package.json file does not exist.')
                self.direct_dep = False
        except Exception as e:
            logger.warning(f'Cannot print direct/transitive dependency: {e}')
            self.direct_dep = False

    def parse_oss_information(self, f_name):
        with open(f_name, 'r', encoding='utf8') as json_file:
            json_data = json.load(json_file)

        sheet_list = []
        comment = ''
        _licenses = 'licenses'
        _repository = 'repository'
        _private = 'private'

        keys = [key for key in json_data]

        for i in range(0, len(keys)):
            d = json_data.get(keys[i - 1])
            oss_init_name = d['name']
            oss_name = self.package_manager_name + ":" + oss_init_name

            if d[_licenses]:
                license_name = d[_licenses]
            else:
                license_name = ''

            oss_version = d['version']
            package_path = d['path']

            private_pkg = False
            if _private in d:
                if d[_private]:
                    private_pkg = True

            homepage = self.dn_url + oss_init_name
            dn_loc = f"{self.dn_url}{oss_init_name}/v/{oss_version}"
            purl = get_url_to_purl(dn_loc, self.package_manager_name)
            self.purl_dict[f'{oss_init_name}({oss_version})'] = purl
            if d[_repository]:
                dn_loc = d[_repository]
            elif private_pkg:
                dn_loc = ''

            comment_list = []
            deps_list = []
            if private_pkg:
                homepage = dn_loc
                comment_list.append('private')
            if self.package_name == f'{oss_init_name}({oss_version})':
                comment_list.append('root package')
            elif self.direct_dep and len(self.relation_tree) > 0:
                if f'{oss_init_name}({oss_version})' in self.relation_tree[self.package_name]:
                    comment_list.append('direct')
                else:
                    comment_list.append('transitive')

                if f'{oss_init_name}({oss_version})' in self.relation_tree:
                    rel_items = [f'npm:{ri}' for ri in self.relation_tree[f'{oss_init_name}({oss_version})']]
                    deps_list.extend(rel_items)

            manifest_file_path = os.path.join(package_path, const.SUPPORT_PACKAE.get(self.package_manager_name))
            multi_license, license_comment, multi_flag = check_multi_license(license_name, manifest_file_path)

            comment = ','.join(comment_list)
            deps = ','.join(deps_list)
            if multi_flag:
                comment = f'{comment}, {license_comment}'
                sheet_list.append([purl, oss_name, oss_version, multi_license, dn_loc, homepage,
                                  '', '', comment, deps])
            else:
                license_name = license_name.replace(",", "")
                license_name = check_unknown_license(license_name, manifest_file_path)
                sheet_list.append([purl, oss_name, oss_version, license_name, dn_loc, homepage,
                                  '', '', comment, deps])

        return sheet_list


def check_multi_license(license_name, manifest_file_path):
    multi_license_list = []
    multi_license = ''
    license_comment = ''
    multi_flag = False
    try:
        if isinstance(license_name, list):
            for i in range(0, len(license_name)):
                l_i = license_name[i].replace(",", "")
                multi_license_list.append(check_unknown_license(l_i, manifest_file_path))
            multi_license = ','.join(multi_license_list)
            multi_flag = True
        else:
            if license_name.startswith('(') and license_name.endswith(')'):
                license_name = license_name.lstrip('(').rstrip(')')
                license_comment = license_name
                multi_license = ','.join(re.split(r'OR|AND', license_name))
                multi_flag = True
    except Exception as e:
        multi_license = license_name
        logger.warning(f'Fail to parse multi license in npm: {e}')

    return multi_license, license_comment, multi_flag


def check_unknown_license(license_name, manifest_file_path):
    if license_name.endswith('*'):
        license_name = license_name[:-1]

    if license_name == 'UNKNOWN':
        try:
            with open(manifest_file_path, 'r') as f:
                json_f = json.load(f)
                for key in json_f.keys():
                    if key == 'license':
                        license_name = json_f[key]
                        break
        except Exception as e:
            logging.warning(f"Cannot check unknown license: {e}")

    return license_name

==> fosslight_dependency-3.15.5.tar.gz: fosslight_dependency-3.15.5/src/fosslight_dependency/package_manager/Npm.py <==
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 LG Electronics Inc.
# SPDX-License-Identifier: Apache-2.0

import os
import logging
import subprocess
import json
import shutil
import re
import fosslight_util.constant as constant
import fosslight_dependency.constant as const
from fosslight_dependency._package_manager import PackageManager, get_url_to_purl

logger = logging.getLogger(constant.LOGGER_NAME)
node_modules = 'node_modules'


class Npm(PackageManager):
    package_manager_name = const.NPM

    dn_url = 'https://www.npmjs.com/package/'
    input_file_name = 'tmp_npm_license_output.json'
    tmp_custom_json = 'custom.json'
    flag_tmp_node_modules = False

    def __init__(self, input_dir, output_dir):
        super().__init__(self.package_manager_name, self.dn_url, input_dir, output_dir)

    def __del__(self):
        if os.path.isfile(self.input_file_name):
            os.remove(self.input_file_name)
        if self.flag_tmp_node_modules:
            shutil.rmtree(node_modules, ignore_errors=True)
        if os.path.exists(self.tmp_custom_json):
            os.remove(self.tmp_custom_json)

    def run_plugin(self):
        ret = self.start_license_checker()
        return ret

    def start_license_checker(self):
        ret = True
        license_checker_cmd = f'license-checker --production --json --out {self.input_file_name}'
        custom_path_option = ' --customPath '
        npm_install_cmd = 'npm install --production'

        if os.path.isdir(node_modules) != 1:
            logger.info(f"node_modules directory is not existed. So it executes '{npm_install_cmd}'.")
            self.flag_tmp_node_modules = True
            cmd_ret = subprocess.call(npm_install_cmd, shell=True)
            if cmd_ret != 0:
                logger.error(f"{npm_install_cmd} returns an error")
                return False

        # customized json file for obtaining specific items with license-checker
        self.make_custom_json(self.tmp_custom_json)

        cmd = license_checker_cmd + custom_path_option + self.tmp_custom_json
        cmd_ret = subprocess.call(cmd, shell=True)
        if cmd_ret != 0:
            logger.error(f"It returns the error: {cmd}")
            logger.error("Please check if the license-checker is installed.(sudo npm install -g license-checker)")
            ret = False
        else:
            self.append_input_package_list_file(self.input_file_name)
        if os.path.exists(self.tmp_custom_json):
            os.remove(self.tmp_custom_json)

        return ret

    def make_custom_json(self, tmp_custom_json):
        with open(tmp_custom_json, 'w', encoding='utf8') as custom:
            custom.write(
                "{\n\t\"name\": \"\",\n\t\"version\": \"\",\n\t\"licenses\": \"\",\n\t\"repository\": \
                \"\",\n\t\"url\": \"\",\n\t\"copyright\": \"\",\n\t\"licenseText\": \"\"\n}\n".encode().decode("utf-8"))

    def parse_rel_dependencies(self, rel_name, rel_ver, rel_dependencies):
        _dependencies = 'dependencies'
        _version = 'version'
        _peer = 'peerMissing'

        for rel_dep_name in rel_dependencies.keys():
            # Optional, non-installed dependencies are listed as empty objects
            if rel_dependencies[rel_dep_name] == {}:
                continue
            if _peer in rel_dependencies[rel_dep_name]:
                if rel_dependencies[rel_dep_name][_peer]:
                    continue
            if f'{rel_name}({rel_ver})' not in self.relation_tree:
                self.relation_tree[f'{rel_name}({rel_ver})'] = []
            elif f'{rel_dep_name}({rel_dependencies[rel_dep_name][_version]})' in self.relation_tree[f'{rel_name}({rel_ver})']:
                continue
            self.relation_tree[f'{rel_name}({rel_ver})'].append(f'{rel_dep_name}({rel_dependencies[rel_dep_name][_version]})')
            if _dependencies in rel_dependencies[rel_dep_name]:
                self.parse_rel_dependencies(rel_dep_name, rel_dependencies[rel_dep_name][_version],
                                            rel_dependencies[rel_dep_name][_dependencies])

    def parse_transitive_relationship(self):
        _dependencies = 'dependencies'
        _version = 'version'
        _name = 'name'
        ret = True
        err_msg = ''

        cmd = 'npm ls -a --omit=dev --json -s'
        result = subprocess.run(cmd, shell=True, capture_output=True, text=True, encoding='utf-8')
        rel_tree = result.stdout
        if rel_tree is None:
            logger.error(f"It returns the error: {cmd}")
            logger.error(f"No output for {cmd}")
            ret = False
        if ret:
            if result.returncode == 1:
                logger.warning(f"'{cmd}' returns error code: {result.stderr}")

            try:
                rel_json = json.loads(rel_tree)
                if len(rel_json) < 1:
                    ret = False
                else:
                    self.package_name = f'{rel_json[_name]}({rel_json[_version]})'
                    if _dependencies in rel_json:
                        self.parse_rel_dependencies(rel_json[_name], rel_json[_version], rel_json[_dependencies])
            except Exception as e:
                ret = False
                err_msg = e
        return ret, err_msg

    def parse_direct_dependencies(self):
        if not self.direct_dep:
            return
        try:
            if os.path.isfile(const.SUPPORT_PACKAE.get(self.package_manager_name)):
                ret, err_msg = self.parse_transitive_relationship()
                if not ret:
                    self.direct_dep = False
                    logger.warning(f'It cannot print direct/transitive dependency: {err_msg}')
            else:
                logger.info('Direct/transitive support is not possible because the package.json file does not exist.')
                self.direct_dep = False
        except Exception as e:
            logger.warning(f'Cannot print direct/transitive dependency: {e}')
            self.direct_dep = False

    def parse_oss_information(self, f_name):
        with open(f_name, 'r', encoding='utf8') as json_file:
            json_data = json.load(json_file)

        sheet_list = []
        comment = ''
        _licenses = 'licenses'
        _repository = 'repository'
        _private = 'private'

        keys = [key for key in json_data]

        for i in range(0, len(keys)):
            d = json_data.get(keys[i - 1])
            oss_init_name = d['name']
            oss_name = self.package_manager_name + ":" + oss_init_name

            if d[_licenses]:
                license_name = d[_licenses]
            else:
                license_name = ''

            oss_version = d['version']
            package_path = d['path']

            private_pkg = False
            if _private in d:
                if d[_private]:
                    private_pkg = True

            homepage = self.dn_url + oss_init_name
            dn_loc = f"{self.dn_url}{oss_init_name}/v/{oss_version}"
            purl = get_url_to_purl(dn_loc, self.package_manager_name)
            self.purl_dict[f'{oss_init_name}({oss_version})'] = purl
            if d[_repository]:
                dn_loc = d[_repository]
            elif private_pkg:
                dn_loc = ''

            comment_list = []
            deps_list = []
            if private_pkg:
                homepage = dn_loc
                comment_list.append('private')
            if self.package_name == f'{oss_init_name}({oss_version})':
                comment_list.append('root package')
            elif self.direct_dep and len(self.relation_tree) > 0:
                if f'{oss_init_name}({oss_version})' in self.relation_tree[self.package_name]:
                    comment_list.append('direct')
                else:
                    comment_list.append('transitive')

                if f'{oss_init_name}({oss_version})' in self.relation_tree:
                    deps_list.extend(self.relation_tree[f'{oss_init_name}({oss_version})'])

            manifest_file_path = os.path.join(package_path, const.SUPPORT_PACKAE.get(self.package_manager_name))
            multi_license, license_comment, multi_flag = check_multi_license(license_name, manifest_file_path)

            comment = ','.join(comment_list)
            if multi_flag:
                comment = f'{comment}, {license_comment}'
                license_name = multi_license
            else:
                license_name = license_name.replace(",", "")
                license_name = check_unknown_license(license_name, manifest_file_path)
            sheet_list.append([purl, oss_name, oss_version, license_name, dn_loc, homepage,
                              '', '', comment, deps_list])
        sheet_list = self.change_dep_to_purl(sheet_list)
        return sheet_list


def check_multi_license(license_name, manifest_file_path):
    multi_license_list = []
    multi_license = ''
    license_comment = ''
    multi_flag = False
    try:
        if isinstance(license_name, list):
            for i in range(0, len(license_name)):
                l_i = license_name[i].replace(",", "")
                multi_license_list.append(check_unknown_license(l_i, manifest_file_path))
            multi_license = ','.join(multi_license_list)
            multi_flag = True
        else:
            if license_name.startswith('(') and license_name.endswith(')'):
                license_name = license_name.lstrip('(').rstrip(')')
                license_comment = license_name
                multi_license = ','.join(re.split(r'OR|AND', license_name))
                multi_flag = True
    except Exception as e:
        multi_license = license_name
        logger.warning(f'Fail to parse multi license in npm: {e}')

    return multi_license, license_comment, multi_flag


def check_unknown_license(license_name, manifest_file_path):
    if license_name.endswith('*'):
        license_name = license_name[:-1]

    if license_name == 'UNKNOWN':
        try:
            with open(manifest_file_path, 'r') as f:
                json_f = json.load(f)
                for key in json_f.keys():
                    if key == 'license':
                        license_name = json_f[key]
                        break
        except Exception as e:
            logging.warning(f"Cannot check unknown license: {e}")

    return license_name
